'From Squeak3.7beta of ''1 April 2004'' [latest update: #5868] on 3 April 2004 at 9:16:12 am'!
"Change Set:		NoAssocDict-ajh
Date:			3 April 2004
Author:			Anthony Hannan

Identity dictionary that uses no associations and a new largeHash function.  Dictionaries of this type cannot be used in compiler bindings.  It is meant for large dictionaries although it does not help too much.  The largeHash function is calculated using the object's identityHash and the class's identityHash, thus it only distributes decently in heterogeneous collections."!

IdentityDictionary subclass: #NoAssocIdentityDictionary
	instanceVariableNames: 'keys'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Collections-Unordered'!

!NoAssocIdentityDictionary commentStamp: '<historical>' prior: 0!
Save space by not using an association for each binding, instead use a single keys array and a single values array.  Also, to get a broader distribution, we hash on #largeHash, which is a combination of identityHash and class's identityHash.!


!ProtoObject methodsFor: 'comparing' stamp: 'ajh 10/12/2001 21:35'!
largeHash
	"Make a larger identity hash using my class's identity hash"

	^ (self identityHash bitShift: 12) + (self class identityHash bitXor: self identityHash)! !


!Array methodsFor: 'comparing' stamp: 'ajh 10/12/2001 21:47'!
largeHash
	"Make a larger identity hash incorporating my size (which does not change)"

	^ (self identityHash bitShift: 12) + (self identityHash bitXor: self size)! !


!CompiledMethod methodsFor: 'comparing' stamp: 'ajh 10/12/2001 21:48'!
largeHash
	"Make a larger identity hash incorporating my header"

	^ (self identityHash bitShift: 12) + (self identityHash bitXor: self header)! !


!NoAssocIdentityDictionary methodsFor: 'accessing' stamp: 'ajh 10/12/2001 15:43'!
add: anAssociation
	^ self at: anAssociation key put: anAssociation value! !

!NoAssocIdentityDictionary methodsFor: 'accessing' stamp: 'ajh 4/3/2004 07:34'!
at: key ifAbsent: aBlock

	| index |
	index _ self findElementOrNil: key.
	(keys at: index) == nil ifTrue: [ ^ aBlock value ].
	^ array at: index! !

!NoAssocIdentityDictionary methodsFor: 'accessing' stamp: 'ajh 4/3/2004 07:34'!
at: key put: value
	"Set the value at key to be value."

	| index |
	index _ self findElementOrNil: key.
	(keys at: index) == nil ifTrue: [
		tally _ tally + 1.
		keys at: index put: key].
	array at: index put: value.
	self fullCheck.
	^ value! !

!NoAssocIdentityDictionary methodsFor: 'accessing' stamp: 'ajh 4/3/2004 07:36'!
includesKey: aSymbol
	"This override assumes that pointsTo is a fast primitive"

	aSymbol ifNil: [^ false].
	^ keys pointsTo: aSymbol! !

!NoAssocIdentityDictionary methodsFor: 'accessing' stamp: 'ajh 4/3/2004 07:37'!
keyAtIdentityValue: value ifAbsent: exceptionBlock
	"Answer the key whose value equals the argument, value. If there is
	none, answer the result of evaluating exceptionBlock."
	
	| theKey |
	1 to: keys size do:
		[:index |
		value == (array at: index)
			ifTrue:
				[(theKey _ keys at: index) == nil
					ifFalse: [^ theKey]]].
	^ exceptionBlock value! !

!NoAssocIdentityDictionary methodsFor: 'accessing' stamp: 'ajh 4/3/2004 07:37'!
keyAtValue: value ifAbsent: exceptionBlock
	"Answer the key whose value equals the argument, value. If there is
	none, answer the result of evaluating exceptionBlock."
	
	| theKey |
	1 to: keys size do:
		[:index |
		value = (array at: index)
			ifTrue:
				[(theKey _ keys at: index) == nil
					ifFalse: [^ theKey]]].
	^ exceptionBlock value! !

!NoAssocIdentityDictionary methodsFor: 'accessing' stamp: 'ajh 9/6/2002 02:53'!
keysSortedSafely
	"Do not sort, because my collection is usually too large"

	^ self keys! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 07:51'!
copy
	^ self shallowCopy withKeys: keys shallowCopy withArray: array shallowCopy! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 08:13'!
grow

	| oldKeys oldValues key index |
	oldKeys _ keys.
	oldValues _ array.
	keys _ Array new: array size + self growSize.
	array _ Array new: keys size.
	1 to: oldKeys size do: [:i |
		(key _ oldKeys at: i) ifNotNil: [
			index _ self findElementOrNil: key.
			keys at: index put: key.
			array at: index put: (oldValues at: i)]].
! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 08:30'!
init: n
	"Initialize array to an array size of n"

	super init: n.
	keys _ Array new: n.
! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 08:09'!
keyAt: index

	^ keys at: index! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 07:58'!
noCheckAdd: assoc

	| index |
	index _ self findElementOrNil: assoc key.
	keys at: index put: assoc key.
	array at: index put: assoc value.
	tally _ tally + 1! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 08:14'!
rehash

	| oldKeys oldValues key index |
	oldKeys _ keys.
	oldValues _ array.
	keys _ Array new: array size.
	array _ Array new: keys size.
	1 to: oldKeys size do: [:i |
		(key _ oldKeys at: i) ifNotNil: [
			index _ self findElementOrNil: key.
			keys at: index put: key.
			array at: index put: (oldValues at: i)]].
! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 08:19'!
scanFor: anObject
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	| element start finish |
	start _ (anObject largeHash \\ array size) + 1.
	finish _ array size.

	"Search from (hash mod size) to the end."
	start to: finish do:
		[:index | ((element _ keys at: index) == nil or: [element == anObject])
			ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ keys at: index) == nil or: [element == anObject])
			ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 08:18'!
swap: oneIndex with: otherIndex

	keys swap: oneIndex with: otherIndex.
	super swap: oneIndex with: otherIndex.
! !

!NoAssocIdentityDictionary methodsFor: 'private' stamp: 'ajh 4/3/2004 07:51'!
withKeys: keysArray withArray: anArray
	"private -- for use only in copy"
	
	keys _ keysArray.
	array _ anArray! !

!NoAssocIdentityDictionary methodsFor: 'enumeration' stamp: 'ajh 4/3/2004 08:20'!
associationsDo: aBlock

	| key |
	tally = 0 ifTrue: [^ self].
	1 to: keys size do:
		[:i | (key _ keys at: i) == nil ifFalse:
			[aBlock value: (Association key: key
									value: (array at: i))]]! !

!NoAssocIdentityDictionary methodsFor: 'enumeration' stamp: 'ajh 4/3/2004 08:25'!
do: aBlock

	self valuesDo: aBlock! !

!NoAssocIdentityDictionary methodsFor: 'enumeration' stamp: 'ajh 4/3/2004 08:22'!
keysAndValuesDo: aBlock 
	"Enumerate the receiver with all the keys and values passed to the block"
	
	| key |
	tally = 0 ifTrue: [^ self].
	1 to: keys size do:
		[:i | (key _ keys at: i) == nil ifFalse:
			[aBlock value: key value: (array at: i)]
		]! !

!NoAssocIdentityDictionary methodsFor: 'enumeration' stamp: 'ajh 4/3/2004 08:22'!
keysDo: aBlock

	| key |
	tally = 0 ifTrue: [^ self].
	1 to: keys size do:
		[:i | (key _ keys at: i) == nil
			ifFalse: [aBlock value: key]]! !

!NoAssocIdentityDictionary methodsFor: 'enumeration' stamp: 'ajh 4/3/2004 08:24'!
valuesDo: aBlock

	tally = 0 ifTrue: [^ self].
	1 to: keys size do:
		[:i | (keys at: i) == nil
			ifFalse: [aBlock value: (array at: i)]]! !

!NoAssocIdentityDictionary methodsFor: 'removing' stamp: 'ajh 4/3/2004 08:26'!
removeKey: key ifAbsent: aBlock

	| index element |
	index _ self findElementOrNil: key.
	(keys at: index) == nil ifTrue: [ ^ aBlock value ].
	element _ array at: index.
	array at: index put: nil.
	keys at: index put: nil.
	tally _ tally - 1.
	self fixCollisionsFrom: index.
	^ element! !


!SmallInteger methodsFor: 'comparing' stamp: 'ajh 10/12/2001 21:48'!
largeHash

	^ self! !
