'From Squeak3.7beta of ''1 April 2004'' [latest update: #5868] on 3 April 2004 at 10:34:06 am'!
"Change Set:		RehashingSystemTracer-svp
Date:			3 April 2004
Author:			Stephen Pair

A subclass of SystemTracer2 that allows you to easily change object identityHashs.  See RehashingSystemTracer class comment.  SystemTracer2-ajh (3 April 2004) must be filed-in before this."!

SystemTracer2 subclass: #RehashingSystemTracer
	instanceVariableNames: 'lastHash hashMap'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'System-Tracing'!

!RehashingSystemTracer commentStamp: 'svp 8/23/2002 20:28' prior: 0!
I am just like SystemTracer2, except that I write new identity hashes for objects.  I rehash some special sets (like MethodDictionaries) prior to writing the clone image, and the clone image will rehash all objects when it starts up.

My instance variables:

lastHash - the lastHash that was assigned...I use the same algorithm to generate hashes as the VM
hashMap - A dictionary of all assigned identity hashes 

Subclasses can extend me if they need to write new images with modified hash values.!


!ProtoObject methodsFor: '*systemtracer2' stamp: 'svp 8/22/2002 21:45'!
rehashWithoutBecome
	"Do nothing.  Here so sending this to a Set does not have to do a time consuming respondsTo:"

	^self rehash! !


!RehashingSystemTracer methodsFor: 'as yet unclassified' stamp: 'svp 8/22/2002 08:59'!
hashMap

	^hashMap! !

!RehashingSystemTracer methodsFor: 'as yet unclassified' stamp: 'ajh 4/3/2004 10:30'!
newHashFor: object
	"This method is here for compatibility with #hashMappedBy methods"

	^self identityHashFor: object! !

!RehashingSystemTracer methodsFor: 'as yet unclassified' stamp: 'svp 8/20/2002 16:33'!
newObjectHash
	"Answer a new 16-bit pseudo-random number for use as an identity hash."

	lastHash := 13849 + (27181 * lastHash) bitAnd: 65535.
	^ lastHash
! !

!RehashingSystemTracer methodsFor: 'as yet unclassified' stamp: 'svp 8/22/2002 22:19'!
specialSets
	"The sets listed here get rehashed using the newly assigned identity hash
	values as the clone is written.  This list probably includes more items than 
	absolutely necessary, but some of these items (especially the method dictionaries)
	are absolutely necessary to get the clone image up and running.  On startup, 
	the clone image will rehash all objects to ensure that all hashed structures are 
	properly arranged.  Rehashing is done in this two step process to retain maximum
	flexibility (if we tried to do everything up front, we would need to update the 
	tracer every time a new type of Set is added with a different hashing and scanning
	algorithm)."

	| params |
	^{
		(Symbol classPool at: #SymbolTable).
		(Symbol classPool at: #NewSymbols).
		Smalltalk.
		(Preferences classPool at: #DictionaryOfPreferences).
		(params := Preferences classPool at: #Parameters).
		(params at: #windowColors).
		(StandardScriptingSystem classPool at: #FormDictionary).
		(StandardScriptingSystem classPool at: #HelpStrings).
	}, MethodDictionary allInstances
! !

!RehashingSystemTracer methodsFor: 'clone startup' stamp: 'svp 8/22/2002 22:01'!
clonePreStartup

	Smalltalk rehashEntireImage! !

!RehashingSystemTracer methodsFor: 'object encoding' stamp: 'svp 8/22/2002 15:49'!
identityHashFor: object

	^hashMap at: object ifAbsentPut: 
		[self newObjectHash bitAnd: self hashMask]
! !

!RehashingSystemTracer methodsFor: 'object encoding' stamp: 'svp 8/21/2002 12:46'!
lastHash

	^lastHash! !

!RehashingSystemTracer methodsFor: 'object replacement' stamp: 'svp 8/22/2002 22:13'!
initSwapMapPostShutdown
	"Subclasses may want to add converted objects to swap map before starting trace but after shuting down extra processes"

	| clone |
	self specialSets do: 
		[ :ea |
		clone := ea cloneUsingHashesFrom: self.
		swapMap at: ea put: clone.
		swapMap at: ea array put: clone array].
! !

!RehashingSystemTracer methodsFor: 'private' stamp: 'svp 12/6/2002 12:55'!
initialize: numObjects

	super initialize: numObjects.
	hashMap := LargeIdentityDict new: numObjects.
	lastHash := 999.
! !


!Set methodsFor: 'private' stamp: 'svp 8/23/2002 09:24'!
compare: element with: object

	^element = object! !

!Set methodsFor: 'private' stamp: 'svp 8/23/2002 09:48'!
findElementOrNil: anObject hash: hash
	"Answer the index of a first slot containing either a nil (indicating an empty slot) or an element that matches the given object. Answer the index of that slot or zero. Fail if neither a match nor an empty slot is found."

	| index |

	index _ self scanFor: anObject hash: hash.
	index > 0 ifTrue: [^index].

	"Bad scene.  Neither have we found a matching element
	nor even an empty slot.  No hashed set is ever supposed to get
	completely full."
	self error: 'There is no free space in this set!!'.! !

!Set methodsFor: 'private' stamp: 'svp 8/23/2002 09:47'!
scanFor: anObject hash: hash
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."

	^self 
		scanFor: anObject 
		startingAt: (hash \\ array size) + 1
! !

!Set methodsFor: 'private' stamp: 'svp 8/23/2002 10:09'!
scanFor: anObject startingAt: start
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	| element |

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | ((element _ array at: index) == nil or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ array at: index) == nil or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !


!IdentityDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:26'!
cloneUsingHashesFrom: tracer

	| object newDict |
	newDict := self shallowCopy.
	newDict withArray: (Array new: array size).
	1 to: array size do:
		[ :i |
		object := array at: i.
		object ifNotNil:
			[newDict
				noCheckAdd: object
				hash: (tracer identityHashFor: object key)]].
	^newDict! !

!IdentityDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 18:36'!
noCheckAdd: anObject hash: hash
	"Variant to allow the hash to be specified
	by the caller.  8/23/02 svp"

	array at: (self findElementOrNil: anObject key hash: hash) put: anObject.
	tally _ tally + 1! !


!MethodDictionary methodsFor: 'private' stamp: 'svp 8/23/2002 17:32'!
scanFor: anObject startingAt: start
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements."
	| element |

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | ((element _ self basicAt: index) == nil or: 
			[element == anObject])
				ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ self basicAt: index) == nil or: 
			[element == anObject])
				ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !

!MethodDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:27'!
cloneUsingHashesFrom: tracer

	| newDict key |
	newDict := self class basicNew: self basicSize.
	newDict init: array size.
	1 to: self basicSize do: 
		[ :i |
		key := self basicAt: i.
		key ifNotNil: 
			[newDict 
				noCheckAt: key 
				put: (array at: i) 
				hash: (tracer identityHashFor: key)]].
	^newDict! !

!MethodDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:26'!
noCheckAt: key put: value hash: identityHash
	"Set the value at key to be value."
	| index |
	index _ self findElementOrNil: key hash: identityHash.
	(self basicAt: index) == nil
		ifTrue: 
			[tally _ tally + 1.
			self basicAt: index put: key].
	array at: index put: value.
	^ value! !


!String methodsFor: '*systemtracer2' stamp: 'svp 8/22/2002 22:45'!
hashMappedBy: map
	"My hash is independent of my oop."

	"While this statement is technically correct, it is misleading.  My hash does
	depend on the identity hash of my species, therefore it does need special care
	when cloning an image with new identity hashes - svp"

	^String stringHash: self initialHash: (self species hashMappedBy: map)! !


!Symbol methodsFor: '*systemtracer2' stamp: 'svp 11/15/2002 16:14'!
hashMappedBy: map
	"** DELETEME ** Overridden here to work with DVS packaging mechanism."

	^super hashMappedBy: map! !


!SystemDictionary methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 00:11'!
rehashEntireImage
	"Note, we cannot use allObjectsDo: because some objects like to swap identities
	during a rehash...this means that allObjectsDo: will not find all objects.  Also, this
	algorithm delays become operations to the very end where we can do them all at once. -svp"

	| objectCount allObjects numMdicts oldObjs newObjs newObject |

	numMdicts := MethodDictionary allInstances size.
	oldObjs := OrderedCollection new: numMdicts.
	newObjs := OrderedCollection new: numMdicts.

	"Create an array big enough to hold all objects in memory"
	objectCount := 0.
	self allObjectsDo: [ :ea | objectCount := objectCount + 1].
	allObjects := Array new: objectCount + 500.

	"Now fill the array"
	objectCount := 0.
	self allObjectsDo: [ :ea | allObjects at: (objectCount := objectCount + 1) put: ea ].

	'Rehashing the image: ', objectCount printString, ' objects' 
		displayProgressAt: Display center
		from: 1 to: objectCount
		during: [:bar | 
			objectCount := 0.
			allObjects do: [ :ea |
				bar value: (objectCount := objectCount + 1).
				newObject := ea rehashWithoutBecome.
				(newObject ~~ ea) ifTrue: 
					[oldObjs add: ea.
					newObjs add: newObject].
			].
			oldObjs asArray elementsExchangeIdentityWith: newObjs asArray.
		].! !


!WeakSet methodsFor: 'private' stamp: 'svp 8/23/2002 17:32'!
scanFor: anObject startingAt: start
	"Scan the key array for the first slot containing either a nil (indicating an empty slot) or an element that matches anObject. Answer the index of that slot or zero if no slot is found. This method will be overridden in various subclasses that have different interpretations for matching elements"

	| element |

	"Search from (hash mod size) to the end."
	start to: array size do:
		[:index | ((element _ array at: index) == flag or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	"Search from 1 to where we started."
	1 to: start-1 do:
		[:index | ((element _ array at: index) == flag or: 
			[self compare: element with: anObject])
				ifTrue: [^ index ]].

	^ 0  "No match AND no empty slot"! !

!WeakSet methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:26'!
cloneUsingHashesFrom: tracer

	| newSet object |
	newSet := self class basicNew init: array size.
	1 to: array size do:
		[ :i |
		object := array at: i.
		(object == flag or: [object == nil]) ifFalse:
			[newSet 
				noCheckAdd: object
				hash: (object hashMappedBy: tracer)]].
	^newSet! !

!WeakSet methodsFor: '*systemtracer2' stamp: 'svp 8/23/2002 16:27'!
noCheckAdd: anObject hash: hash
	array at: (self findElementOrNil: anObject hash: hash) put: anObject.
	tally _ tally + 1! !

