'From Squeak3.7beta of ''1 April 2004'' [latest update: #5868] on 3 April 2004 at 10:23:47 am'!
"Change Set:		SystemTracer2-ajh
Date:			3 April 2004
Author:			Anthony Hannan

A more flexible and reliable SystemTracer.  See SystemTracer2 class comment.  File-in RehashingSystemTracer-svp after this if you want to change identityHashes."!

Stream subclass: #PositionableStream
	instanceVariableNames: 'collection position readLimit '
	classVariableNames: 'IntBuffer '
	poolDictionaries: ''
	category: 'Collections-Streams'!
Object subclass: #SystemTracer2
	instanceVariableNames: 'swapMap oopMap traceQueue byteStream nextChunk displayBox nilOop'
	classVariableNames: 'ImageHeaderSize LargeIdentityDict'
	poolDictionaries: ''
	category: 'System-Tracing'!

!SystemTracer2 commentStamp: 'ajh 4/3/2004 09:48' prior: 0!
This is a rewrite of SystemTracer.  It makes it easier to convert images, although it requires more space.

Use it by subclassing it and overriding desired methods in:
	'object relacement' to create replacement objects for desired objects (swapMap);
	'object encoding' to specify bit representation of objects;
	'clone startup' to initialize the new image on first startup.

The default is to trace and write the system in V3 image format.

To run the tracer do "YourTracerSubclass writeImage: 'newImageFile.image'".  The .changes file will be copied as well.  The tracer requires a lot of space to run since it is keeps an oopMap for every object in the system, and a swapMap for every object replacement.  And because these maps are very large and Squeak has a poor identityHash function, the tracer runs slowly.


Structure:

 swapMap	IdentityDictionary (oldObject -> newObject)
			When an object is visited, swapMap is checked to see if it has a replacement, and if so the new object is traced and written instead of the original.  You can add to the swapMap in initSwapMapPreShutdown or initSwapMapPostShutdown before tracing starts, or you can add to it during tracing in convert:pointer:field:.  The tracer automatically adds itself and the active process to swapMap so it own execution won't be traced (pvtWriteImageConverted).

 oopMap	IdentityDictionary (object -> oop)
			Maps objects to their new oops.  The tracer checks this in case a visited object has already been reached.

 traceQueue  OrderedCollection (object)
			Queue of objects that have been reached by the tracer and assigned a new oop, but have not yet been written and traced themselves.  The image is traced by adding specialObjectArray to the queue, then repeatedly removing the first object and tracing it until the queue is empty.  An object is traced by adding all its field values (including the class field) that haven't already been reached (oopMap) to the queue.

 byteStream  FileStream
			The target image file that gets written to during tracing.  Oops equal stream positions.

 nextChunk  Integer
			Oop (minus header size) assigned to the next object reached by the tracer.

 displayBox  Rectangle
			Bounds of Squeak window; written in image header.

 nilOop 		Integer
			Oop of new nil object.  Need here since nil can't be added as a key to oopMap dictionary.

Class Vars:

ImageHeaderSize		Integer
					number of bytes in header of image file.

LargeIdentityDict	IdentityDictionary class
					The tracer uses this class for its large maps, so link your favorite large dictionary class to this.
!


!ProtoObject methodsFor: 'system primitives' stamp: 'ajh 7/12/2001 19:58'!
basicSize
	"Primitive. Answer the number of indexable variables in the receiver. 
	This value is the same as the largest legal subscript. Essential. Do not 
	override in any subclass. See Object documentation whatIsAPrimitive."

	<primitive: 62>
	"The number of indexable fields of fixed-length objects is 0"
	^ 0! !

!ProtoObject methodsFor: 'system tracing' stamp: 'ajh 8/20/2002 18:11'!
allFieldsWithIndex: block collect: sequenceableCollectionClass
	"Evaluate block against each of my pointer fields with index, and collect the results in an instance of sequenceableCollectionClass"

	| f v results |
	f _ self class instSize.
	v _ self basicSize.
	results _ sequenceableCollectionClass new: f + v.
	1 to: f do: [:j | results at: j put: (block value: (self instVarAt: j) value: j)].
	1 to: v do: [:j | results at: f + j put: (block value: (self basicAt: j) value: f + j)].
	^ results! !

!ProtoObject methodsFor: 'system tracing' stamp: 'ajh 4/3/2004 07:07'!
forSystemTracer: swapMap
	"May want to create a new object and add it to swapMap under self.  Be sure to check that self has not already been added to swapMap, since self may receive this message multiple times if it is referenced by multiple objects.  Return the new object that is to substitute for self, or self is there is no substitute.  It is ok for the substitute to have pointers to objects that still need to be converted.  They will be converted when the substitute is traced by the system tracer.  Alternatively, you can convert them right away by sending forSystemTracer: to them.
	If you know only certain classes will be converted then it may be faster to override this method in subclasses that get converted and make this method just return self, but be sure you don't neglect objects that are added to the swapMap by the tracer itself (see references to swapMap in SystemTracer2)"

	^ swapMap at: self ifAbsent: [self]! !

!ProtoObject methodsFor: 'system tracing' stamp: 'ajh 8/20/2002 23:27'!
forSystemTracer: swapMap pointer: pointer field: index
	"May want to create a new object and add it to swapMap under self.
	 pointer is the object pointing to obj via its field at index.  You may want to convert obj to different objects depending on who is pointing to it.  If this is the case you will want to remember the multiple mappings in swapMap.  You can do this by keeping another map inside the swapMap keyed at a special symbol"

	^ self forSystemTracer: swapMap! !

!ProtoObject methodsFor: 'system tracing' stamp: 'ajh 10/15/2001 11:20'!
numFields
	"Return my number of fields (bytes or words)"

	^ self class instSize + self basicSize! !


!Behavior methodsFor: 'system tracing' stamp: 'ajh 1/18/2002 20:12'!
indexIfCompactForConversion
	"subclasses may override this to change cc index during conversion.  You must update the compact classes array yourself"

	^ self indexIfCompact! !


!CompiledMethod methodsFor: 'system tracing' stamp: 'ajh 8/20/2002 18:13'!
allFieldsWithIndex: block collect: sequenceableCollectionClass
	"Evaluate block against each of my pointer fields with index, and collect the results in an instance of sequenceableCollectionClass"

	| results |
	results _ sequenceableCollectionClass new: 1 + self numLiterals.
	1 to: 1 + self numLiterals do: [:j | 
		results at: j put: (block value: (self objectAt: j) value: j)].
	^ results! !


!CompiledMethod class methodsFor: 'class initialization' stamp: 'ajh 7/7/2001 16:37'!
fullFrameSize

	^ LargeFrame! !


!ContextPart methodsFor: 'system tracing' stamp: 'ajh 8/20/2002 18:14'!
allFieldsWithIndex: block collect: sequenceableCollectionClass
	"Evaluate block against each of my pointer fields with index, and collect the results in an instance of sequenceableCollectionClass.  Include entire stack capacity"

	| nilResults j |
	nilResults _ sequenceableCollectionClass new: self frameSize - self basicSize.
	j _ 0.
	self basicSize + 1 to: self frameSize do: [:i |
		nilResults at: (j _ j + 1) put: (block value: nil value: i)].
	^ (super allFieldsWithIndex: block collect: sequenceableCollectionClass), nilResults! !

!ContextPart methodsFor: 'system tracing' stamp: 'ajh 10/15/2001 11:18'!
numFields
	"Return my number of fields (bytes or words)"

	^ self class instSize + self frameSize ! !


!BlockContext methodsFor: 'private' stamp: 'ajh 7/7/2001 16:44'!
frameSize

	^ home ifNil: [CompiledMethod fullFrameSize] ifNotNil: [self method frameSize]! !


!MethodContext methodsFor: 'private' stamp: 'ajh 7/7/2001 16:44'!
frameSize

	^ method ifNil: [CompiledMethod fullFrameSize] ifNotNil: [self method frameSize]! !


!PositionableStream methodsFor: 'nonhomogeneous accessing' stamp: 'ajh 10/13/2001 15:11'!
nextLongPut: positiveInt
	"Write positiveInt in the next four bytes preserving the current endianess"

	self nextPutAll: (IntBuffer at: 1 put: positiveInt; yourself)
! !


!PositionableStream class methodsFor: 'instance creation' stamp: 'ajh 10/13/2001 15:12'!
initialize

	IntBuffer _ WordArray new: 1.! !


!SystemTracer2 methodsFor: 'run' stamp: 'ajh 4/3/2004 10:23'!
writeImage: imageFileName
	"Trace all object in the image and write them out to imageFileName.  See class comment for details."
	"SystemTracer2 writeImage: 'clone.image'"

	self initialize: self class numObjects.
	self initSwapMapPreShutdown.
	Smalltalk garbageCollect.
	(SourceFiles at: 2) ifNotNil:[
		Smalltalk saveChangesInFileNamed: (Smalltalk fullNameForChangesNamed: imageFileName)].
	displayBox _ Display boundingBox.
	byteStream _ FileStream fileNamed: imageFileName.
	byteStream binary.
	[	self processShutdown.
		self pvtWriteImageConverted.
		"new clone image will start up from here"
	] on: Error do: [:ex |
		self processStartup.
		ex pass.
	].
	"isStartingClone will be true in new image"
	self isStartingClone ifFalse: [byteStream close].
	self isStartingClone ifTrue: [self clonePreStartup].
	self processStartup.
	self isStartingClone ifTrue: [self clonePostStartup].
	self isStartingClone ifFalse: [self inform: 'Done tracing!!'].
! !

!SystemTracer2 methodsFor: 'object replacement' stamp: 'ajh 4/3/2004 06:38'!
convert: obj pointer: pointer field: index
	"May want to substitute a new object for obj.  And you may want to substitute diferent objects depending on the pointer.  pointer is the object pointing to obj via its field at index.
	Delegate to obj so it can decide if it wants to convert and how.  Make sure they check the swapMap in case a substitute has already been created for them in initSwapMapPreShutdown or initSwapMapPostShutdown or pvtWriteImageConverted.  Each SystemTracer subclass should probably use a different convert message that looks like 'forSubclassName:...' so each subclass can have its own set of conversion methods.  You can model them after the forSystemTracer:... methods."

	^ obj forSystemTracer: swapMap pointer: pointer field: index! !

!SystemTracer2 methodsFor: 'object replacement' stamp: 'ajh 4/3/2004 04:29'!
initSwapMapPostShutdown
	"Subclasses may want to add converted objects to swap map before starting trace but after shuting down extra processes"
! !

!SystemTracer2 methodsFor: 'object replacement' stamp: 'ajh 4/3/2004 04:32'!
initSwapMapPreShutdown
	"Subclasses may want to add converted objects to swapMap before starting trace and before shuting down extra processes"
! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 4/3/2004 09:52'!
hashMask

	^ 16rFFF! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 4/3/2004 09:52'!
headersFor: class classOop: classOop hash: hash numFields: length
	"Return header words for new object.
	 Three possibilities:
		{length, class, baseHeader}
		{class, baseHeader}
		{baseHeader}"

	| header3 header2 header1 cix sizeFld |
	"3 gc bits"
	header1 _ 0.  "Mark, old, dirty"

	"Add hash"
	header1 _ header1 bitShift: 12.	"next fld is 12 bits"
	header1 _ header1 + (hash bitAnd: self hashMask).

	"Add compact class index or class header"
	header1 _ header1 bitShift: 5.
	cix _ class indexIfCompactForConversion.	"0 means need full word"
	header2 _ classOop.
	header1 _ header1 + (cix bitAnd: 16r1F).

	"Add object format"
	header1 _ header1 bitShift: 4.
	header1 _ header1 + (class instSpec +
		(class isWords ifTrue: [0] ifFalse: [3 - (length + 3 bitAnd: 3)])).
						"For byte objects: size = wordSize - spec.lowBits"

	"Add size in long words (includes base header)"
	header1 _ header1 bitShift: 6.
	sizeFld _ (class isBytes ifTrue: [length + 3 // 4] ifFalse: [length]) + 1.
	sizeFld > 16r3F
		ifTrue: [header3 _ sizeFld bitShift: 2.  sizeFld _ 0]
		ifFalse: [header3 _ 0].
	header1 _ header1 + sizeFld.

	header1 _ header1 bitShift: 2.

	header3 > 0 ifTrue: ["3-word: type=0"
		^ WordArray with: header3 + 0 with: header2 + 0 with: header1 + 0].
	cix = 0 ifTrue: [ "2-word: type=1"
		^ WordArray with: header2 + 1 with: header1 + 1].
	"1-word: type=3"
	^ WordArray with: header1 + 3! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 8/20/2002 19:53'!
identityHashFor: object
	"object is the new object if the original was converted.  If you change the identity hash, you will want to rehash all objects in cloneStartUp.  However, if you change the identity hash of selectors (Symbols), then you should convert MethodDictionaries and selectors up front, before/during tracing"

	^ object identityHash! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 8/20/2002 16:50'!
integerOopOf: value

	value < 0
		ifTrue: [^ ((16r80000000 + value) << 1) + 1]
		ifFalse: [^ (value << 1) + 1]! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 4/3/2004 04:26'!
lastHash

	^ Object new identityHash! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 5/28/2002 22:48'!
lengthAndHeaderSizeFor: class numFields: n
	"Return instance word length (including base header), and instance header size (including base header).  This method should be consistent with headersFor:classOop:hash:numFields:"

	| cix size |
	size _ (class isBytes ifTrue: [n + 3 // 4] ifFalse: [n]) + 1.
			"size in long words, incl hdr0"
	cix _ class indexIfCompactForConversion.	"0 means need full word"
	size > 16r3F ifTrue: [^ {size. 3}].
	cix = 0 ifTrue: [^ {size. 2}].
	^ {size. 1}! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 4/3/2004 04:40'!
versionLetter
	"Image format major version number"

	^ $A! !

!SystemTracer2 methodsFor: 'object encoding' stamp: 'ajh 4/3/2004 04:40'!
versionNumber
	"Image format minor version number"

	^ 2! !

!SystemTracer2 methodsFor: 'clone startup' stamp: 'ajh 8/22/2002 11:14'!
clonePostStartup
	"This will be executed when the new clone starts up, but after processStartup.  Subclasses may want to open a greeting window or something"! !

!SystemTracer2 methodsFor: 'clone startup' stamp: 'ajh 8/22/2002 11:15'!
clonePreStartup
	"This will be executed right away when the new clone starts up, before processStartup.  Subclasses may want to rehash all objects or something"! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 4/3/2004 05:01'!
initialize: numObjects

	swapMap _ LargeIdentityDict new: 4000.
	oopMap _ LargeIdentityDict new: numObjects.
	traceQueue _ OrderedCollection new: numObjects // 4.
! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/20/2002 16:38'!
isStartingClone
	"byteStream will be nil in new image"

	^ byteStream == nil! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/20/2002 16:31'!
lengthAndHeaderSizeFor: object
	"Return two sizes in an array:
	 1. object's size in words including my base header,
	 2. my header size in words including base header"

	^ self lengthAndHeaderSizeFor: object class numFields: object numFields! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/22/2002 11:17'!
processShutdown

	(Object respondsTo: #flushDependents) ifTrue: [Object flushDependents].
	(Object respondsTo: #flushEvents) ifTrue: [Object flushEvents].
	Smalltalk processShutDownList: false.  "false means not quiting"
	Cursor write show.
! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/22/2002 11:17'!
processStartup

	Cursor normal show.
	Smalltalk setGCParameters.
	self isStartingClone ifTrue: [Smalltalk clearExternalObjects].
	Smalltalk processStartUpList: self isStartingClone.
	self isStartingClone ifTrue: [
		Smalltalk setPlatformPreferences.
		Smalltalk readDocumentFile].
	Smalltalk isMorphic ifTrue: [SystemWindow wakeUpTopWindowUponStartup].
! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 4/3/2004 03:16'!
pvtTraceImage

	'Tracing and writing ', (oopMap capacity * 3//4) printString, ' objects' 
		displayProgressAt: Display center
		from: ImageHeaderSize to: self class imageSize
		during: [:bar |  | lengthAndHeaderSize object oop |

			"Add true and false in front so the VM can assume it will never move (see Interpreter>>interpret)"
			self reserve: true.
			self reserve: false.

			"Need to add nil separately because it can't be added as a key to oopMap"
			lengthAndHeaderSize _ self lengthAndHeaderSizeFor: nil.
			nilOop _ nextChunk + ((lengthAndHeaderSize second - 1) * 4).
			nextChunk _ nilOop + (lengthAndHeaderSize first * 4).
			self writeAndTrace: nil.

			"Add specialObjectArray next.  Use new array in swap map if present.  This is done explicitly here so swap map does not have to be tested on every Array write"
			object _ swapMap at: Smalltalk specialObjectsArray ifAbsent: [Smalltalk specialObjectsArray].
			lengthAndHeaderSize _ self lengthAndHeaderSizeFor: object.
			oop _ nextChunk + ((lengthAndHeaderSize second - 1) * 4).
			nextChunk _ oop + (lengthAndHeaderSize first * 4).
			oopMap at: Smalltalk specialObjectsArray put: oop.
			oopMap at: object put: oop.
			self writeAndTrace: object.

			"Write and trace rest of image"
			[traceQueue isEmpty] whileFalse: [
				bar value: (self writeAndTrace: traceQueue removeFirst)].
		].

	^ nil! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 4/3/2004 04:26'!
pvtWriteFileHeader

	| specialObjectsOop |
	specialObjectsOop _ oopMap at: Smalltalk specialObjectsArray.
	byteStream position: 0.
	byteStream nextLongPut: (self versionLetter asciiValue * 100) + self versionNumber.
	byteStream nextLongPut: ImageHeaderSize.
	byteStream nextLongPut: nextChunk - ImageHeaderSize.
	byteStream nextLongPut: ImageHeaderSize.  "start of memory"
	byteStream nextLongPut: specialObjectsOop.
	byteStream nextLongPut: self lastHash.
	byteStream nextLongPut: displayBox width * 16r10000 + displayBox height.
	byteStream nextLongPut: 0.  "fullScreenFlag"
	byteStream nextLongPut: Smalltalk extraVMMemory.
	byteStream padTo: ImageHeaderSize put: 0.
! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/20/2002 23:14'!
pvtWriteImageConverted

	"Map self to empty and thisContext to nil so they don't get traced"
	swapMap at: self put: self class basicNew.
	swapMap at: thisContext put: nil.

	"Make new image start up in sender (writeImageFile:) right after this call.  Push nil since the first instr after the call will be pop"
	thisContext sender push: nil.
	swapMap at: Processor activeProcess put: (Process forContext: thisContext sender priority: Processor activePriority).

	"May add more to swapMap before starting trace"
	self initSwapMapPostShutdown.
	Smalltalk garbageCollect.

	"Trace starting from specialObjectArray.  Also, subclasses may convert each object while tracing by implementing #convert:pointer:field"
	nextChunk _ ImageHeaderSize.
	self pvtTraceImage.

	self pvtWriteFileHeader.

	thisContext sender pop.  "pop nil pushed above for clone image"
	^ nil! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/20/2002 17:40'!
reserve: object
	"Add object to write queue and return its oop"

	| lengthAndHeaderSize objectSize hdrSize oop |
	"Calculate oop and reserve space (advance nextChunk)"
	lengthAndHeaderSize _ self lengthAndHeaderSizeFor: object.
	objectSize _ lengthAndHeaderSize at: 1.
	hdrSize _ lengthAndHeaderSize at: 2.
	oop _ nextChunk + ((hdrSize - 1) * 4).
	nextChunk _ oop + (objectSize * 4).

	"Add object to fill queue and remember its oop"
	oopMap at: object put: oop.
	traceQueue addLast: object.

	^ oop! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/20/2002 19:24'!
reserve: obj from: pointer field: index
	"Add obj to write queue and return its oop"

	| object |
	obj class = SmallInteger ifTrue: [^ self integerOopOf: obj].

	"Return oop if already visited obj"
	object _ self convert: obj pointer: pointer field: index.
	object ifNil: [^ nilOop].
	oopMap at: object ifPresent: [:itsOop | ^ itsOop].

	^ self reserve: object! !

!SystemTracer2 methodsFor: 'private' stamp: 'ajh 8/20/2002 19:28'!
writeAndTrace: object
	"Write the image representation of me on byteStream at my oop, and add my not-yet-seen fields to traceQueue."

	| lengthAndHeaderSize oop byteSize byteBuffer |
	lengthAndHeaderSize _ self lengthAndHeaderSizeFor: object.
	oop _ oopMap at: object ifAbsent: [object isNil ifTrue: [nilOop] ifFalse: [self halt: 'oop should have been added to oopMap when object was add to traceQueue']].

	"Write header"
	byteStream position: oop - ((lengthAndHeaderSize second - 1) * 4).
	byteStream nextPutAll: (self
		headersFor: object class
		classOop: (self reserve: object class from: object field: -1 "class field")
		hash: (self identityHashFor: object)
		numFields: object numFields
	).

	"Write fields"
	(object class isPointers or: [object class == CompiledMethod]) ifTrue: [
		byteStream nextPutAll: (object
			allFieldsWithIndex: [:val :i | self reserve: val from: object field: i]
			collect: WordArray).
	] ifFalse: [  "isBits"
		byteStream nextPutAll: object.
		object class isBytes ifTrue: [  "fill bytes unused in last word"
			1 to: 3 - (object size + 3 bitAnd: 3) do: [:i | byteStream nextPut: 0]].
	].
	object class == CompiledMethod ifTrue: [
		byteSize _ object basicSize - object initialPC + 1.
		byteBuffer _ ByteArray new: byteSize.
		byteBuffer replaceFrom: 1 to: byteSize with: object startingAt: object initialPC.
		"intermediate ByteArray needed since byteStream (a file) cannot putAll from a subclass of ByteArray (CompiledMethod)"
		byteStream nextPutAll: byteBuffer.
		"fill bytes unused in last word"
		1 to: 3 - (object size + 3 bitAnd: 3) do: [:i | byteStream nextPut: 0].
	].

	"Double check that fields take up expected size"
	byteStream position = (oop + (lengthAndHeaderSize first * 4)) ifFalse: [
		self error: 'object size discrepency'].

	^ oop! !


!SystemTracer2 class methodsFor: 'image' stamp: 'ajh 7/9/2001 19:08'!
imageSize

	^ Smalltalk vmParameterAt: 2! !

!SystemTracer2 class methodsFor: 'image' stamp: 'ajh 7/3/2001 00:53'!
numObjects
	"SystemTracer2 numObjects"

	| count object |
	count _ 0.
	object _ 0 someObject.
	[object == 0] whileFalse: [
		count _ count + 1.
		object _ object nextObject].
	^ count! !

!SystemTracer2 class methodsFor: 'initialize' stamp: 'ajh 4/3/2004 07:29'!
initialize

	ImageHeaderSize _ 64.  "16 words"
	LargeIdentityDict _ Smalltalk at: #LargeIdentityDictionary
		ifAbsent: [Smalltalk at: #NoAssocIdentityDictionary
			ifAbsent: [IdentityDictionary]].
! !

!SystemTracer2 class methodsFor: 'initialize' stamp: 'ajh 4/3/2004 08:48'!
useLargeDictClass: aIdentityDictionaryClass

	LargeIdentityDict _ aIdentityDictionaryClass! !

!SystemTracer2 class methodsFor: 'run' stamp: 'ajh 4/3/2004 08:48'!
writeImage: imageFileName
	"Trace all object in the image and write them out to imageFileName.  See class comment for details."
	"SystemTracer2 writeImage: 'clone.image'"

	^ self new writeImage: imageFileName! !


SystemTracer2 initialize!
PositionableStream initialize!
